 # -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint
import os
import json
import csv
import logging
import math

import parse
import proteins as parse_proteins
import peptidemass

"""
Parser for Morpheus files
"""


logger = logging.getLogger('morpheus')


def read_tsv_iter(tsv_txt):
  """
  Reads a title top TSV file, converts all keys to lower case
  and returns a list of dictionaries.
  """
  f = open(tsv_txt, "UR")
  titles = [w.lower() for w in parse.split_tab(f.readline())]
  results = []
  for line in f:
    group = {}
    for key, val in zip(titles, parse.split_tab(line)):
      group[key] = parse.parse_string(val)
    yield group
    del group


def read_modification_dict(modifications_tsv):
  result = {}
  for entry in parse.read_tsv(modifications_tsv):
    # pprint(entry)
    key = entry['description']
    mass = float(entry['monoisotopic mass shift (da)'])
    if 'residue' in entry:
      aa = entry['residue']
    elif 'amino acid residue' in entry:
      aa = entry['amino acid residue']
    else:
      aa = 'n/a'
    if aa != 'n/a':
      mass += peptidemass.aa_monoisotopic_mass[aa]
    result[key] = mass
  return result


def parse_peptide(text, modification_dict):
  seq = text.split('.')[1]
  chars = []
  modifications = []
  in_modification = False
  mod_str = ''
  for c in seq:
    if in_modification:
      if c == ']' or c == ')':
        in_modification = False
        i = len(chars) - 1
        if mod_str not in modification_dict:
          log.debug('Warning: modification {} unknown'.format(mod_str))
          continue
        modification = {
            'i':i, 
            'mass': modification_dict[mod_str],
        }
        modifications.append(modification)
      else:
        mod_str += c
    else:
      if c == '[' or c == '(':
        in_modification = True
        mod_str = ''
      else:
        chars.append(c)
  return ''.join(chars), modifications


def get_first(s, delimiter='/'):
  if not isinstance(s, str):
    return s
  elif delimiter not in s:
    return s
  return s.split(delimiter)[0].strip()


class DictListWriter():
  def __init__(self, is_debug, fname):
    self.fname = fname
    self.is_debug = is_debug
    if self.is_debug:
      logger.debug('Dumping dict list to ' + self.fname)
      self.file = open(self.fname, 'w')
      self.file.write('[\n')

  def dump_dict(self, data_dict):
    if self.is_debug:
      pprint(data_dict, stream=self.file)
      self.file.write(',\n')

  def close(self):
    if self.is_debug:
      self.file.write(']\n')
      self.file.close()


def get_proteins(protein_groups_fname, psm_fname, modifications_fname=None):
  is_debug = logger.root.level <= logging.DEBUG
  dump_dir = os.path.dirname(protein_groups_fname)

  if modifications_fname is not None:
    modification_table = read_modification_dict(modifications_fname)
  else:
    modification_table = {}
  
  proteins = {}
  dict_dump_writer = DictListWriter(is_debug, os.path.join(dump_dir, 'protein_groups.dump'))
  for i_group, protein_group in enumerate(read_tsv_iter(protein_groups_fname)):
    descriptions = protein_group['protein description'].split(' / ')
    coverage_str = str(protein_group['protein sequence coverage (%)'])
    if ';' in coverage_str:
      coverage =  float(get_first(coverage_str, ';'))
    else:
      coverage =  float(get_first(coverage_str, '/'))
    seqs = protein_group['protein sequence'].split('/')
    seqids = [desc.split()[0] for desc in descriptions]
    for seqid in seqids:
      if seqid in proteins:
        logger.warning("Different protein groups claim same first seqid", seqid)
    protein = {
      'description': descriptions[0],
      'sequence': seqs[0],
      'other_sequences': seqs[1:],
      'attr': {
        'coverage': parse.round_decimal(coverage, 4),
        'morpheus-score': parse.round_decimal(protein_group['summed morpheus score'], 4),
        'i_group': i_group,
        'other_seqids': seqids[1:],
        'seqid': seqids[0],
      },
      'sources': [{ 'peptides':[] }]
    }
    proteins[seqids[0]] = protein
    dict_dump_writer.dump_dict(protein_group)
  dict_dump_writer.close()

  protein_by_seqid = {}
  for seqid in proteins:
    protein = proteins[seqid]
    protein_by_seqid[seqid] = protein
    for alt_seqid in protein['attr']['other_seqids']:
      protein_by_seqid[alt_seqid] = protein

  dict_dump_writer = DictListWriter(is_debug, os.path.join(dump_dir, 'peptides.dump'))
  n_peptide = 0
  n_peptide_matched = 0
  for src_peptide in read_tsv_iter(psm_fname):
    dict_dump_writer.dump_dict(src_peptide)
    descriptions = src_peptide['protein description'].split(' / ')
    peptide_seqids = [d.split()[0] for d in descriptions]
    protein = None
    for peptide_seqid in peptide_seqids:
      if peptide_seqid in protein_by_seqid:
        protein = protein_by_seqid[peptide_seqid]
        break
    n_peptide += 1
    if protein is None:
      continue
    n_peptide_matched += 1
    sequence = protein['sequence']
    extracted_peptide_sequence, modifications = parse_peptide(
        src_peptide['peptide sequence'],
        modification_table)
    peptide_sequence = src_peptide['base peptide sequence']
    if extracted_peptide_sequence != peptide_sequence:
      logger.warning("Peptide sequences don't match: " + src_peptide['peptide sequence'] + " " + extracted_peptide_sequence + " " + peptide_sequence)
    i = sequence.find(peptide_sequence)
    if i < 0:
      logger.warning(peptide_sequence + ' not found in ' + protein['attr']['seqid'])
      continue
    q_value = float(src_peptide['q-value (%)'])
    if 'scan number' in src_peptide:
      scan_id = src_peptide['scan number']
    elif 'spectrum number' in src_peptide:
      scan_id = src_peptide['spectrum number']
    else:
      scan_id = ''
    if 'retention time (min)' in src_peptide:
      time = parse.round_decimal(src_peptide['retention time (min)'], 4)
    elif 'retention time (minutes)' in src_peptide:
      time = parse.round_decimal(src_peptide['retention time (minutes)'], 4)
    else:
      time = ''

    peptide = {
      'sequence': peptide_sequence,
      'attr': {
        'scan_id': scan_id, 
        'retention_time': time,
        'morpheus_score': parse.round_decimal(src_peptide['morpheus score'], 4),
        'mass': parse.round_decimal(src_peptide['precursor mass (da)'], 4),
        'mass_diff': parse.round_decimal(src_peptide['precursor mass error (da)'], 4),
        'm/z': parse.round_decimal(src_peptide['precursor m/z'], 4),
        'source': parse.basename(src_peptide['filename']),
        'q_value': q_value,
      },
      'intensity': 1.0 - q_value/100.0,
      'i': i,
    }
    if modifications:
      for modification in modifications:
        modification['mass'] = parse.round_decimal(modification['mass'], 4)
      peptide['attr']['modifications'] = modifications

    protein['sources'][0]['peptides'].append(peptide)

  dict_dump_writer.close()

  dump = os.path.join(dump_dir, 'proteins.dump')
  if logger.root.level <= logging.DEBUG:
    logger.debug('Dumping proteins data structure to ' + dump)
    parse.save_data_dict(proteins, dump)

  logger.info("Assigned {}/{} of PSMs.tsv to protein_groups.tsv".format(n_peptide_matched, n_peptide))

  return proteins


if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG)
  proteins = get_proteins(
      '../example/morpheus/OK20130822_MPProtomap_KO1.protein_groups.tsv',
      '../example/morpheus/OK20130822_MPProtomap_KO1.PSMs.tsv',
      '../example/morpheus/modifications.tsv')





 # -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint

import math
import os
import json
import copy
import glob
import shutil
from pprint import pprint
import logging


import fasta

import uniprot


logger = logging.getLogger('proteins')


this_dir = os.path.abspath(os.path.dirname(__file__))


def new_protein(seqid):
  return {
    'attr': { 'seqid':seqid, 'other_seqids':[] },
    'description': '',
    'sequence': '',
    'sources': [{'peptides': [], }],
  }


def new_peptide(peptide_sequence):
  return {
    'sequence': peptide_sequence,
    'intensity': 1.0,
    'mask': 0.0,
    'attr': {}
  }


def peptide_iter(proteins):
  for seqid in proteins:
    protein = proteins[seqid]
    for source in protein['sources']:
      for peptide in source['peptides']:
        yield seqid, peptide


def determine_unique_peptides(proteins):
  seqids_by_seq = {}
  for seqid, peptide in peptide_iter(proteins):
    seq = peptide['sequence']
    if seq not in seqids_by_seq:
      seqids_by_seq[seq] = set()
    seqids_by_seq[seq].add(seqid)
  for seqid, peptide in peptide_iter(proteins):
    seq = peptide['sequence']
    n_seqid = len(seqids_by_seq[seq])
    peptide['attr']['is_unique'] = (n_seqid == 1)


def check_missing_fields(proteins):
  for seqid, peptide in peptide_iter(proteins):
    if 'modifications' not in peptide['attr']:
      peptide['attr']['modifications'] = []
    if 'intensity' not in peptide:
      peptide['intensity'] = 1
    if 'mask' not in peptide:
      peptide['mask'] = 0.0


def count_peptides(proteins):
  for seqid in proteins.keys():
    protein = proteins[seqid]
    n_peptide = 0
    n_slice_populated = 0
    n_unique_peptide = 0
    for source in protein['sources']:
      peptides = source['peptides']
      unique_peptides = [p for p in peptides if p['attr']['is_unique']]
      n_unique_peptide += len(unique_peptides)
      n_peptide += len(peptides)
      if len(peptides) > 0:
        n_slice_populated += 1
    protein['attr']['n_slice_populated'] = n_slice_populated
    protein['attr']['n_unique_peptide'] = n_unique_peptide
    protein['attr']['n_peptide'] = n_peptide
    if n_peptide == 0:
      del proteins[seqid]
      continue


def find_peptide_proteins(proteins):
  n_source = len(proteins.values()[0]['sources'])
  seqids_by_sequence = {}
  for seqid in proteins:
    protein = proteins[seqid]
    for i_source in range(n_source):
      source = protein['sources'][i_source]
      peptides = source['peptides']
      for peptide in peptides:
        sequence = peptide['sequence']
        if sequence not in seqids_by_sequence:
          seqids_by_sequence[sequence] = set()
        seqids_by_sequence[sequence].add(seqid)
  for seqid in proteins:
    protein = proteins[seqid]
    for i_source in range(n_source):
      protein = proteins[seqid]
      source = protein['sources'][i_source]
      peptides = source['peptides']
      for peptide in peptides:
        sequence = peptide['sequence']
        for test_seqid in seqids_by_sequence[sequence]:
          if seqid != test_seqid:
            if 'other_seqids' not in peptide['attr']:
              peptide['attr']['other_seqids'] = []
            peptide['attr']['other_seqids'].append(test_seqid)
            logger.debug('{} in {} also found in {}'.format(sequence, seqid, test_seqid))


def change_seqids_in_proteins(proteins, clean_seqid):
  seqids = proteins.keys()
  for seqid in seqids:
    new_seqid = clean_seqid(seqid)
    if 'attr' in proteins[seqid]:
      if 'other_seqids' in proteins[seqid]['attr']:
        other_seqids = proteins[seqid]['attr']['other_seqids']
        other_seqids = map(clean_seqid, other_seqids)
        proteins[seqid]['attr']['other_seqids'] = other_seqids
    if seqid != new_seqid:
      proteins[new_seqid] = proteins[seqid]
      del proteins[seqid]
      if 'attr' in proteins[new_seqid]:
        proteins[new_seqid]['attr']['seqid'] = new_seqid


def calculate_peptide_positions(proteins, iso_leu_isomerism=False):
  for seqid in proteins:
    protein = proteins[seqid]
    protein_sequence = protein['sequence']
    if iso_leu_isomerism:
      protein_sequence = protein_sequence.replace("L", "I")
    for source in protein['sources']:
      peptides = source['peptides']
      for i_peptide in reversed(range(len(peptides))):
        peptide = peptides[i_peptide]
        peptide_sequence = peptide['sequence']
        if iso_leu_isomerism:
          peptide_sequence = peptide_sequence.replace("L", "I")
        i = protein_sequence.find(peptide_sequence)
        if i < 0:
          logger.debug("'{}' not found in {}".format(peptide_sequence, seqid))
          del peptides[i_peptide]
          continue
        peptide['i'] = i 
      for peptide in peptides:
        if 'i' not in peptide:
          pprint(peptide)
          raise ValueError


def load_fastas_into_proteins(
    proteins, fastas, clean_seqid=None, iso_leu_isomerism=False):
  if clean_seqid:
    change_seqids_in_proteins(proteins, clean_seqid)
    change_seqids_in_proteins(fastas, clean_seqid)
  for seqid in proteins.keys():
    protein = proteins[seqid]
    if seqid not in fastas:
      logger.debug("%s of proteins not found in fastas" % seqid)
      del proteins[seqid]
      continue
    protein_sequence = fastas[seqid]['sequence']
    protein['description'] = fastas[seqid]['description']
    protein['sequence'] = protein_sequence
    protein['attr']['length'] = len(protein_sequence)
  calculate_peptide_positions(proteins, iso_leu_isomerism)


def load_sequences_from_uniprot(proteins, clean_seqid=None, cache_basename=None):
  if clean_seqid:
    change_seqids_in_proteins(proteins, clean_seqid)
  seqids = []
  for seqid in proteins:
    seqids.append(seqid)
    if 'other_seqids' in proteins[seqid]['attr']:
      seqids.extend(proteins[seqid]['attr']['other_seqids'])
  uniprot_data = uniprot.get_metadata_with_some_seqid_conversions(seqids, cache_basename)
  load_fastas_into_proteins(proteins, uniprot_data)
  if cache_basename:
    uniprot.write_fasta(cache_basename+'.fasta', uniprot_data, uniprot_data.keys())


def load_fasta_db_into_proteins(
    proteins, fasta_db, clean_seqid=None, iso_leu_isomerism=False):
  seqids, fastas = fasta.read_fasta(fasta_db)
  load_fastas_into_proteins(proteins, fastas, clean_seqid, iso_leu_isomerism)


def merge_two_proteins(proteins1, proteins2):
  """
  Merges two proteins structures. In particular, it grafts the
  'sources' together, treating the sources in each proteins as 
  distinct, and maintaing the order.
  """
  if len(proteins1) == 0:
    return proteins2
  seqid = proteins1.keys()[0]
  n_source1 = len(proteins1[seqid]['sources'])
  seqid = proteins2.keys()[0]
  n_source2 = len(proteins2[seqid]['sources'])
  for seqid in proteins2:
    protein2 = proteins2[seqid]
    sources2 = protein2['sources']
    if seqid in proteins1:
      protein1 = proteins1[seqid]
      protein1['sources'].extend(sources2)
    else:
      sources1 = [{'peptides': []} for i in range(n_source1)]
      combined_sources = sources1 + sources2
      proteins1[seqid] = protein2
      proteins1[seqid]['sources'] = combined_sources
  for seqid in proteins1:
    peptides = proteins1[seqid]['sources']
    if len(peptides) == n_source1:
      peptides.extend([{'peptides': []} for i in range(n_source2)])
  return proteins1


def save_data_js(data, js_fname):
  f = open(js_fname, 'w')
  f.write('var data = \n')
  f.write(json.dumps(data, indent=None))
  f.close()


def transfer_files(in_dir, out_dir):
  if not os.path.isdir(out_dir):
    os.makedirs(out_dir)
  for src in glob.glob(os.path.join(in_dir, '*')):
    dst = os.path.join(out_dir, os.path.basename(src))
    shutil.copy(src, dst)


def make_graphical_comparison_visualisation(data, out_dir):
  # sanity checks
  proteins = data['proteins']
  determine_unique_peptides(proteins)
  count_peptides(proteins)
  check_missing_fields(proteins)
  find_peptide_proteins(proteins)
  for seqid, protein in proteins.items():
    for source in protein['sources']:
      peptides = source['peptides']
      peptides.sort(key=lambda peptide: len(peptide['sequence']))
      peptides.sort(key=lambda peptide: peptide['i'])
  if 'source_labels' not in data:
    data['source_labels'] = []
  if 'color_names' not in data:
    data['color_names'] = ['', '', '']
  if 'mask_labels' not in data:
    data['mask_labels'] = []

  if not os.path.isdir(out_dir):
    os.makedirs(out_dir)

  save_data_js(data, os.path.join(out_dir, 'data.js'))
  transfer_files(os.path.join(this_dir, 'templates/comparison'), out_dir)
  transfer_files(os.path.join(this_dir, 'templates/js'), os.path.join(out_dir, 'js'))
  index_html = os.path.abspath(os.path.join(out_dir, 'index.html'))
  logger.info('Made peptograph in "' + index_html + '"')


def make_sequence_overview_visualisation(data, out_dir):
  # sanity checks
  proteins = data['proteins']
  determine_unique_peptides(proteins)
  count_peptides(proteins)
  check_missing_fields(proteins)
  find_peptide_proteins(proteins)
  for seqid, protein in proteins.items():
    for source in protein['sources']:
      peptides = source['peptides']
      peptides.sort(key=lambda peptide: len(peptide['sequence']))
      peptides.sort(key=lambda peptide: peptide['i'])
  if 'source_labels' not in data:
    data['source_labels'] = []
  if 'color_names' not in data:
    data['color_names'] = ['', '', '']
  if 'mask_labels' not in data:
    data['mask_labels'] = []

  if not os.path.isdir(out_dir):
    os.makedirs(out_dir)

  save_data_js(data, os.path.join(out_dir, 'data.js'))
  transfer_files(os.path.join(this_dir, 'templates/overview'), out_dir)
  transfer_files(os.path.join(this_dir, 'templates/js'), os.path.join(out_dir, 'js'))
  index_html = os.path.abspath(os.path.join(out_dir, 'index.html'))
  logger.info('Made peptograph in "' + index_html + '"')








 # -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint

import pymzml

def load_mzml(proteins, i_source, mzml, n_peak=50):
  peptide_by_scan_id = {}
  for protein in proteins.values():
    source = protein['sources'][i_source]
    for peptide in source['peptides']:
      scan_id = peptide['attr']['scan_id']
      peptide_by_scan_id[scan_id] = peptide
  for spectrum in pymzml.run.Reader(mzml):
    if spectrum['id'] in peptide_by_scan_id:
      peptide = peptide_by_scan_id[spectrum['id']]
      ions = [(mz, i) for mz, i in spectrum.peaks]
      ions.sort(key=lambda i:-i[1])
      peptide['spectrum'] = ions[:n_peak]





# -*- coding: utf-8 -*-

from __future__ import print_function
from pprint import pprint
import xml.etree.ElementTree as etree
import json
import os
import re
import logging

import parse
import proteins as proteins_module


"""
Parser for X!Tandem XML output.
"""


logger = logging.getLogger('xtandem')


def strip_whitespace(txt):
  result = ""
  for line in txt.splitlines():
    result += ''.join(line.split())
  return result


def parse_scan(top_elem, nsmap):
  scan = {}
  scan.update(parse.parse_attrib(top_elem))

  scan['matches'] = []
  for protein in top_elem.findall('protein'):
    words = protein.attrib['label'].split()
    seqid = words[0]
    description = ' '.join(words[1:])
    peptide_elem = protein.find('peptide')
    sequence = strip_whitespace(peptide_elem.text)
    match = {  
      'seqid': seqid,
      'sequence': sequence,
      'description': description
    }
    domain_elem = peptide_elem.find('domain')
    match.update(parse.parse_attrib(domain_elem))
    scan['matches'].append(match)

  elem = top_elem.find('group[@label="fragment ion mass spectrum"]')

  note = elem.find('note')
  if note:
    scan['Description'] = elem.find('note').text.strip()
  else:
    scan['Description'] = ''

  masses_elem = elem.find(
      'GAML:trace/GAML:Xdata/GAML:values', namespaces=nsmap)
  scan['masses'] = masses_elem.text.strip()

  intensities_elem = elem.find(
      'GAML:trace/GAML:Ydata/GAML:values', namespaces=nsmap)
  scan['intensities'] = intensities_elem.text.strip()

  charge_elem = elem.find(
      'GAML:trace/GAML:attribute[@type="charge"]', namespaces=nsmap)
  scan['charge'] = charge_elem.text.strip()

  mass_elem = elem.find(
      'GAML:trace/GAML:attribute[@type="M+H"]', namespaces=nsmap)
  scan['mass'] = mass_elem.text.strip()

  return scan


def read_xtandem(xtandem_xml):
  fastas = {}
  scans = []
  nsmap = {}
  for event, elem in etree.iterparse(xtandem_xml, events=('end', 'start-ns')):
    if event == 'start-ns':
      nsmap.update({elem})
    if event == 'end':
      if elem.tag == 'group' and elem.attrib['type'] == 'model':
        scan = parse_scan(elem, nsmap)
        yield scan
        elem.clear()


def load_xtandem_into_proteins(proteins, xtandem_fname, i_source, n_peak=50):
  proteins_by_seqid = {}
  for seqid in proteins:
    protein = proteins[seqid]
    proteins_by_seqid[seqid] = protein
    for other_seqid in protein['attr']['other_seqids']:
      proteins_by_seqid[other_seqid] = protein

  fastas = {}

  for scan in read_xtandem(xtandem_fname):
    scan_id = scan['id']
    is_match = False
    for match in scan['matches']:
      seqid = match['seqid']
      if seqid not in proteins_by_seqid:
        continue
      protein = proteins_by_seqid[seqid]
      if seqid not in fastas:
        fastas[seqid] = {
          'sequence': match['sequence'],
          'description': match['description'],
        }
      source = protein['sources'][i_source]    
      for peptide in source['peptides']:
        if scan_id != peptide['attr']['scan_id']:
          continue
        is_match = True
        x_vals = map(float, scan['masses'].split())
        y_vals = map(float, scan['intensities'].split())
        ions = [(x, y) for x, y in zip(x_vals, y_vals)]
        ions.sort(key=lambda i:-i[1])
        peptide['spectrum'] = ions[:n_peak]

  proteins_module.load_fastas_into_proteins(proteins, fastas)
  for seqid in proteins.keys():
    protein = proteins[seqid]
    if 'sequence' not in protein:
      logger.debug("Protein {} not found in x!tandem".format(seqid))
      del proteins[seqid]
      continue
    n_peptide = sum([len(source['peptides']) for source in protein['sources']])
    if n_peptide == 0:
      del proteins[seqid]
      logger.debug("No peptide-spectra matches found in {}".format(seqid))
      continue

  # proteins_module.calculate_peptide_positions(proteins)



if __name__ == "__main__":
  scans, fastas = read('../example/xtandem/Seq23282_E1O1.tandem')
  parse.save_data_dict(scans, '../example/xtandem/scans.dump')
  parse.save_data_dict(fastas, '../example/xtandem/fastas.dump')








# -*- coding: utf-8 -*-

from __future__ import print_function
from pprint import pprint
import re
import os
import json
import ntpath, posixpath, macpath


"""
Utility parsing functions that is useful for
different proteomics data parsers.
"""


# string parsers

float_regex_pattern = r"""
^ 
[-+]? # optional sign
(?:
    (?: \d* \. \d+ ) # .1 .12 .123 etc 9.1 etc 98.1 etc
    |
    (?: \d+ \.? ) # 1. 12. 123. etc 1 12 123 etc
)
# followed by optional exponent part if desired
(?: [Ee] [+-]? \d+ ) ?
$
"""
float_regex = re.compile(float_regex_pattern, re.VERBOSE)


def round_decimal(f, n):
    '''Truncates/pads a float f to n decimal places without rounding'''
    slen = len('%.*f' % (n, f))
    return str(f)[:slen]
    
    
def basename(fname):
  for m in ntpath, posixpath, macpath:
    if m.sep in fname:
      return m.basename(fname)
  return fname


def parse_string(s):
  if re.search(r'^[-+]?\d+$', s):
    return int(s)
  elif float_regex.match(s):
    return float(s)
  else:
    return s


# tsv helper function

def splitter(s, convert_fn=None, delimiter=";"):
  s = str(s).strip()
  if not s:
    return []
  words = s.split(delimiter)
  if convert_fn:
    words = map(convert_fn, words)
  return words


def split_tab(line):
  if line[-1] == '\n':
    line = line[:-1]
  return line.split('\t')


def read_tsv(tsv_txt):
  """
  Reads a title top TSV file, converts all keys to lower case
  and returns a list of dictionaries.
  """
  f = open(tsv_txt, "UR")
  titles = [w.lower() for w in split_tab(f.readline())]
  results = []
  for line in f.readlines():
    group = {}
    for key, val in zip(titles, split_tab(line)):
      group[key] = parse_string(val)
    yield group


# XML helper functions


def fixtag(ns, tag, nsmap):
  return '{' + nsmap[ns] + '}' + tag


def parse_attrib(elem):
  result = {}
  for key, value in elem.attrib.items():
    result[key] = parse_string(value)
  return result


def parse_name_value(elem):
  attrib = parse_attrib(elem)
  return { attrib['name']: attrib['value'] }


def save_data_dict(data, fname):
  with open(fname, 'w') as f:
    pprint(data, stream=f, indent=2)


def cache_result_to_json(fname):
  def actual_decorator(fn):
    def wrapped_fn(*args, **kwargs):
      if not os.path.isfile(fname):
        result = fn(*args, **kwargs)
        json.dump(result, open(fname, 'w'))
      else:
        result = json.load(open(fname))
      return result
    return wrapped_fn
  return actual_decorator


def memoize(fn, fname, force=False):
  if not force and os.path.isfile(fname):
    return eval(open(fname).read())
  else:
    result = fn()
    dirname = os.path.dirname(fname)
    if dirname is not '' and not os.path.isdir(dirname):
      os.makedirs(dirname)
    pprint(result, stream=open(fname, 'w'), width=80)
    return result






# -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint


def read_fasta(fasta_db):
  """
  Returns a lsit of seqids encountered, and a dictionary
  of sequences for each seqid.
  """
  seqids = []
  seqid = None
  proteins = {}
  for line in open(fasta_db):
    if line.startswith(">"):
      words = line.split()
      seqid = words[0][1:]
      description = line[len(words[0]):].lstrip()
      seqids.append(seqid)
      proteins[seqid] = {
        'sequence': '',
        'description': description,
      }
      continue
    if seqid is not None:
      words = line.split()
      if words:
        proteins[seqid]['sequence'] += words[0]
  return seqids, proteins
  



 # -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint
import xml.etree.ElementTree as etree
import json
import logging
import os

import parse
import proteins as parse_proteins

"""
Parsers to read TPP output files: PEPXML and PROTXML.
"""

logger = logging.getLogger('tpp')


def parse_scan(scan_elem, nsmap):
  scan = parse.parse_attrib(scan_elem)
  scan['matches'] = []
  tag = lambda tag_id: parse.fixtag('', tag_id, nsmap)
  for search_elem in scan_elem.findall(parse.fixtag('', "search_result", nsmap)):
    search_hit_elem = search_elem[0] 
    match = parse.parse_attrib(search_hit_elem)
    match['modified_sequence'] = match['peptide']

    match['other_seqids'] = []
    for alt_protein in search_hit_elem.findall(parse.fixtag('', 'alternative_protein', nsmap)):
      match['other_seqids'].append(alt_protein.attrib['protein'])

    match['modifications'] = []
    for modified_elem in search_hit_elem.findall(parse.fixtag('', 'modification_info', nsmap)):
      attr = parse.parse_attrib(modified_elem)
      match['modified_sequence'] = attr['modified_peptide']
      for modification_elem in modified_elem.findall(parse.fixtag('', 'mod_aminoacid_mass', nsmap)):
        attr = parse.parse_attrib(modification_elem)
        attr['i'] = attr['position'] - 1
        del attr['position']
        match['modifications'].append(attr)

    for score_elem in search_hit_elem.findall(tag('search_score')):
      match.update(parse.parse_name_value(score_elem))

    for analysis_elem in search_hit_elem.find(parse.fixtag('', 'analysis_result', nsmap)):
      if analysis_elem.tag == parse.fixtag('', 'peptideprophet_result', nsmap):
        match.update(parse.parse_attrib(analysis_elem))
        for param_elem in analysis_elem[0]:
          match.update(parse.parse_name_value(param_elem))

    scan['matches'].append(match)
  return scan


def parse_peptide_probabilities(elem, nsmap):
  # try with error_point
  error_points = elem.findall(parse.fixtag('', 'error_point', nsmap))
  if len(error_points) == 0:
    charge = 0
    for charge_elem in elem.findall(parse.fixtag('', 'roc_error_data', nsmap)):
      if charge_elem.attrib['charge'] == 'all':
        error_points = charge_elem.findall(parse.fixtag('', 'error_point', nsmap))
        break
  probs = []
  for elem in error_points:
      attrib = parse.parse_attrib(elem)
      probs.append({
        'error': attrib['error'],
        'prob': attrib['min_prob'],
      })
  probs.sort(key=lambda d:d['error'])
  return probs


class PepxmlReader(object):
  def __init__(self, pepxml):
    self.pepxml = pepxml
    self.distribution = None
    self.source_names = []
    self.i_source = None
    self.nsmap = {}
    self.is_debug = logger.root.level <= logging.DEBUG

  def __iter__(self):
    if self.is_debug:
      fname = self.pepxml + '.dump'
      logging.debug('Dumping pepxml reads into ' + fname)
      self.debug_file = open(fname, 'w')
      self.debug_file.write('[\n')
    for event, elem in etree.iterparse(self.pepxml, events=('start', 'end', 'start-ns')):
      if event == 'start-ns':
        self.nsmap.update({elem})
      elif event == 'start':
        if elem.tag == parse.fixtag('', 'msms_run_summary', self.nsmap):
          fname = elem.attrib['base_name']
          self.source_names.append(fname)
          self.i_source = len(self.source_names) - 1
      elif event == 'end':
        if elem.tag == parse.fixtag('', 'spectrum_query', self.nsmap):
          scan = parse_scan(elem, self.nsmap)
          for match in scan['matches']:
            fpe = probability_to_error(self.distribution, match['probability'])
            if fpe is None:
              print("WTF", match['probability'], self.distribution)
            else:
              match['fpe'] = fpe
          if self.i_source is not None:
            scan['source'] = self.source_names[self.i_source]
          if self.is_debug:
            pprint(scan, stream=self.debug_file)
            self.debug_file.write(',\n')
          yield scan
          elem.clear()
        elif elem.tag == parse.fixtag('', 'peptideprophet_summary', self.nsmap):
          self.distribution = parse_peptide_probabilities(elem, self.nsmap)
          if self.distribution[0]['prob'] < 1.0:
            self.distribution.insert(0, {'prob':1.0, 'error':0.0})
          if self.distribution[-1]['prob'] > 0.0:
            self.distribution.append({'prob':0.0, 'error':1.0})
          if self.is_debug:
            fname = self.pepxml + '.distribution.dump'
            pprint(self.distribution, open(fname, 'w'))
          elem.clear()
    if self.is_debug:
      self.debug_file.write(']\n')
      self.debug_file.close()


def make_peptide(pepxml_match, pepxml_scan, source):
  peptide = {
    'sequence': pepxml_match['peptide'],
    'modified_sequence': pepxml_match['modified_sequence'],
    'intensity': pepxml_match['probability'],
    'mask': pepxml_match['fpe'],
    'attr': {
      'pepxml_id': pepxml_scan['index'],
      'scan_id': pepxml_scan['start_scan'],
      'charge': pepxml_scan['assumed_charge'],
      'expect': pepxml_match['expect'],
      'modifications': pepxml_match['modifications'],
      'probability': pepxml_match['probability'],
      'missed_cleavages': pepxml_match['num_missed_cleavages'],
      'mass': pepxml_scan['precursor_neutral_mass'],
      'mass_diff': pepxml_match['massdiff'],
      'source': parse.basename(source),
    }
  }
  def grab_opt(peptide_key, scan_key, source_dict):
    if scan_key in source_dict:
      peptide['attr'][peptide_key] = source_dict[scan_key]
  grab_opt('retention_time', 'retention_time_sec', pepxml_scan)
  grab_opt('score', 'ionscore', pepxml_match)
  grab_opt('homology', 'homologyscore', pepxml_match)
  grab_opt('identity', 'identityscore', pepxml_match)
  peptide['attr']['matched_ions'] = str(pepxml_match['num_matched_ions'])
  peptide['attr']['matched_ions'] += '/'
  peptide['attr']['matched_ions'] += str(pepxml_match['tot_num_ions'])
  return peptide


def parse_protein_probabilities(elem, nsmap):
  probs = []
  for data_point in elem.findall(parse.fixtag('', 'protein_summary_data_filter', nsmap)):
    attrib = parse.parse_attrib(data_point)
    probs.append({
      'error': attrib['false_positive_error_rate'],
      'prob': attrib['min_probability'],
    })
  probs.sort(key=lambda d:d['error'])
  return probs


def parse_protein_group(elem, nsmap):
  group = parse.parse_attrib(elem)
  group['proteins'] = []
  for protein_elem in elem.findall(parse.fixtag('', 'protein', nsmap)):
    protein = parse.parse_attrib(protein_elem)
    protein['group_number'] = group['group_number']

    for parameter_elem in protein_elem.findall(parse.fixtag('', 'parameter', nsmap)):
      key = parameter_elem.attrib['name']
      val = parameter_elem.attrib['value']
      protein[key] = val

    annotation_elem = protein_elem.find(parse.fixtag('', 'annotation', nsmap))
    if annotation_elem is not None:
      protein['description'] = annotation_elem.attrib['protein_description']

    protein['other_seqids'] = []
    for alt_protein in protein_elem.findall(parse.fixtag('', 'indistinguishable_protein', nsmap)):
      protein['other_seqids'].append(alt_protein.attrib['protein_name'])

    protein['other_seqids'] = protein['other_seqids']
    protein['protein_name'] = protein['protein_name']

    protein['peptides'] = []
    n_unique_peptide = 0
    for peptide_elem in protein_elem.findall(parse.fixtag('', 'peptide', nsmap)):
      peptide = parse.parse_attrib(peptide_elem)
      protein['peptides'].append(peptide)
      peptide['modifications'] = []
      peptide['modified_sequence'] = peptide['peptide_sequence']
      for modified_elem in peptide_elem.findall(parse.fixtag('', 'modification_info', nsmap)):
        attr = parse.parse_attrib(modified_elem)
        peptide['modified_sequence'] = attr['modified_peptide']
        for modification_elem in modified_elem.findall(parse.fixtag('', 'mod_aminoacid_mass', nsmap)):
          attr = parse.parse_attrib(modification_elem)
          peptide['modifications'].append(attr)

    group['proteins'].append(protein)
  return group


class ProtxmlReader(object):
  def __init__(self, protxml):
    self.protxml = protxml
    self.distribution = None
    self.nsmap = {}
    self.debug_file = None
    self.is_debug = logger.root.level <= logging.DEBUG
    self.nsmap = {}

  def __iter__(self):
    if self.is_debug:
      fname = self.protxml + '.dump'
      logging.debug('Dumping protxml reads into ' + fname)
      self.debug_file = open(fname, 'w')
      self.debug_file.write('{\n')
    for event, elem in etree.iterparse(self.protxml, events=('end', 'start-ns')):
      if event == 'start-ns':
        self.nsmap.update({elem})
      if event == 'end':
        if elem.tag == parse.fixtag('', 'protein_group', self.nsmap):
          group = parse_protein_group(elem, self.nsmap)
          yield group
          if self.is_debug:
            pprint(group, stream=self.debug_file)
            self.debug_file.write(',\n')
          elem.clear()
        elif elem.tag == parse.fixtag('', 'proteinprophet_details', self.nsmap):
          self.distribution = parse_protein_probabilities(elem, self.nsmap)
          if self.is_debug:
            fname = self.protxml + '.distribution.dump'
            pprint(self.distribution, open(fname, 'w'))
          elem.clear()
    if self.is_debug:
      self.debug_file.write('}\n')
      self.debug_file.close()


def make_protein(protxml_protein):
  protein = {
    'description': protxml_protein['description'],
    'attr': { 
      'group_id': protxml_protein['group_number'],
      'length': protxml_protein['prot_length'],
      'seqid': protxml_protein['protein_name'],
      'sibling': protxml_protein['group_sibling_id'],
      'other_seqids': protxml_protein['other_seqids'],
      'probability': protxml_protein['probability'],
      'percent_coverage': '-',
    },
    'sources': [],
  }
  if 'percent_coverage' in protxml_protein:
    protein['attr']['percent_coverage'] = protxml_protein['percent_coverage']
  protein['protxml_peptides'] = protxml_protein['peptides']
  return protein


def generate_proteins_from_protxml(protxml):
  proteins = {}
  protxml_reader = ProtxmlReader(protxml)
  for protein_group in protxml_reader:
    for protxml_protein in protein_group['proteins']:
      seqid = protxml_protein['protein_name']
      if seqid in proteins:
        logger.warning("%s found in multiple protein groups" % seqid)
      proteins[seqid] = make_protein(protxml_protein)
  protein_probs = protxml_reader.distribution
  return proteins, protein_probs


def filter_proteins(proteins, prob_cutoff):
  for seqid in proteins.keys():
    protein = proteins[seqid]
    if protein['attr']['probability'] < prob_cutoff:
      del proteins[seqid]
      continue


def get_protein_by_seqid(proteins):
  protein_by_seqid = {}
  for seqid in proteins:
    protein = proteins[seqid]
    protein_by_seqid[seqid] = protein
    protein = proteins[seqid]
    for alt_seqid in protein['attr']['other_seqids']:
      protein_by_seqid[alt_seqid] = protein
  return protein_by_seqid


def add_source(proteins):
  for protein in proteins.values():
    protein['sources'].append({'peptides': []})


def load_pepxml(proteins, pepxml_fname, prob_cutoff=None, error_cutoff=None, source_names=[]):
  logging.debug('Peptide error cutoff: {}'.format(error_cutoff))
  protein_by_seqid = get_protein_by_seqid(proteins)
  n_source = len(proteins.values()[0]['sources'])
  pepxml_reader = PepxmlReader(pepxml_fname)
  source_name_indices = {}
  for scan in pepxml_reader:
    for match in scan['matches']:
      if error_cutoff is not None and match['fpe'] > error_cutoff:
        continue
      seqids = [match['protein']] + match['other_seqids']
      for seqid in seqids:
        if seqid not in protein_by_seqid:
          logger.warning('{} from scan {} not found in protxml'.format(seqid, scan['index']))
          continue
        protein = protein_by_seqid[seqid]
        if scan['source'] in source_name_indices:
          i_source = source_name_indices[scan['source']]
        else:
          add_source(proteins)
          n_source += 1
          i_source = n_source-1
          source_name_indices[scan['source']] = i_source
        peptides = protein['sources'][i_source]['peptides']
        scan_id = scan['start_scan']
        scan_ids = [p['attr']['scan_id'] for p in peptides]
        if scan_id not in scan_ids:
          peptide = make_peptide(match, scan, scan['source'])
          for protxml_peptide in protein['protxml_peptides']:
            if protxml_peptide['charge'] == scan['assumed_charge'] and \
                protxml_peptide['modified_sequence'] == match['modified_sequence']:  
              peptide['attr']['probability_protxml'] = protxml_peptide['nsp_adjusted_probability']
              peptide['attr']['is_contributing_evidence'] = protxml_peptide['is_contributing_evidence']
          peptides.append(peptide)
  source_names.extend(pepxml_reader.source_names)


def error_to_probability(distribution, error):
  """
  Given a False-Positive-Error vs. Probability distribution,
  Cacluates the probability for an acceptable FPE.
  """
  fractionate = lambda a0, a1, a: (a-a0)/(a1-a0)
  interpolate = lambda a0, a1, f: a0 + f*(a1-a0)
  n = len(distribution)
  for i in range(1, n):
    error0 = distribution[i-1]['error']
    error1 = distribution[i]['error']
    if error0 <= error < error1:
      prob0 = distribution[i-1]['prob']
      prob1 = distribution[i]['prob']
      f = fractionate(error0, error1, error)
      return interpolate(prob0, prob1, f)
  return None


def probability_to_error(distribution, prob):
  """
  Given a False-Positive-Error vs. Probability distribution,
  Cacluates the probability for an acceptable FPE.
  """
  fractionate = lambda a0, a1, a: (a-a0)/(a1-a0)
  interpolate = lambda a0, a1, f: a0 + f*(a1-a0)
  n = len(distribution)
  for i in range(1, n):
    prob0 = distribution[i-1]['prob']
    prob1 = distribution[i]['prob']
    if prob0 >= prob >= prob1:
      error0 = distribution[i-1]['error']
      error1 = distribution[i]['error']
      f = fractionate(prob0, prob1, prob)
      return interpolate(error0, error1, f)
  return None


def filter_peptides(proteins, probability):
  seqids = proteins.keys()
  for seqid in seqids:
    n_peptide = 0
    for source in proteins[seqid]['sources']:
      peptides = source['peptides']
      for i_peptide in reversed(range(len(peptides))):
        if peptides[i_peptide]['attr']['probability'] < probability:
          del peptides[i_peptide]
        else:
          n_peptide += 1
    if n_peptide == 0:
      del proteins[seqid]


def count_tpp_indep_spectra(proteins):
  for seqid in proteins:
    protein = proteins[seqid]
    n_unique_peptide = 0
    n_unique_spectrum = 0
    unique_sequence_set = set()
    for source in protein['sources']:
      for peptide in source['peptides']:
        sequence = str(peptide['attr']['charge']) + peptide['modified_sequence']
        if 'is_contributing_evidence' in peptide['attr'] and peptide['attr']['is_contributing_evidence'] == 'Y':
          n_unique_spectrum += 1
          unique_sequence_set.add(sequence)
    n_unique_peptide = len(unique_sequence_set)
    protein['attr']['n_indep_spectra'] = n_unique_spectrum 
  # calculate percentages of scans
  n_unique_spectrum_total = 0
  for seqid in proteins:
    protein = proteins[seqid]
    n_unique_spectrum_total += protein['attr']['n_indep_spectra']
  for seqid in proteins:
    protein = proteins[seqid]
    if n_unique_spectrum_total == 0:
      percent = 0
    else:
      percent = 100.0*protein['attr']['n_indep_spectra']/n_unique_spectrum_total
    protein['attr']['percent_indep_spectra'] = float('%.2f' % percent)


def get_proteins_and_sources(
    protxml, pepxmls, peptide_error=0.01, protein_error=0.01):
  """
  Returns a proteins dictionary and list of source names.
  """
  logger.info('Loading protxml ' + protxml)
  proteins, protein_probs = generate_proteins_from_protxml(protxml)

  source_names = []
  for pepxml in pepxmls:
    logger.info('Loading pepxml ' + pepxml)
    load_pepxml(proteins, pepxml, error_cutoff=peptide_error, source_names=source_names)
    
  count_tpp_indep_spectra(proteins)

  probability = error_to_probability(protein_probs, protein_error)
  filter_proteins(proteins, probability)

  if logger.root.level <= logging.DEBUG:
    dump = protxml.replace('prot.xml', 'proteins.dump')
    logger.debug('Dumping protein data structure to ' + dump)
    parse.save_data_dict(proteins, dump)

  return proteins, source_names


if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG)
  proteins, sources = get_proteins_and_sources(
    '../example/tpp/hca-lysate-16.prot.xml', 
    ['../example/tpp/hca-lysate-16.pep.xml'])
  



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) 2013 Bosco Ho
from __future__ import print_function
from pprint import pprint

"""
A peptide mass calculation library.
Given sequences in terms of strings, generates all
the different masses of the peptide depending on the
weighted averages of the different isotypes of each atom.
"""

atom_mass = {
  'H': 1.007825032,
  'O': 15.994914622,
}

ave_atom_mass = {
  'H': 1.00794,
  'O': 15.9994,
}

aa_monoisotopic_mass = {
  'A': 71.037114,
  'B': 110.000000,
  'C': 103.009184,
  'D': 115.026943,
  'E': 129.042593,
  'F': 147.068414,
  'G': 57.021464,
  'H': 137.058912,
  'I': 113.084064,
  'J': 110.000000,
  'K': 128.094963,
  'L': 113.084064,
  'M': 131.040485,
  'N': 114.042927,
  'O': 110.000000,
  'P': 97.052764,
  'Q': 128.058578,
  'R': 156.101111,
  'S': 87.032028,
  'T': 101.047678,
  'U': 110.000000,
  'V': 99.068414,
  'W': 186.079313,
  'X': 110.000000,
  'Y': 163.063329,
  'Z': 110.000000
}

aa_average_mass = {
  'A': 71.0788,
  'B': 110.0000,
  'C': 103.1388,
  'D': 115.0886,
  'E': 129.1155,
  'F': 147.1766,
  'G': 57.0519,
  'H': 137.1411,
  'I': 113.1594,
  'J': 110,
  'K': 128.1741,
  'L': 113.1594,
  'M': 131.1926,
  'N': 114.1038,
  'O': 110,
  'P': 97.1167,
  'Q': 128.1307,
  'R': 156.1875,
  'S': 87.0782,
  'T': 101.1051,
  'U': 110,
  'V': 99.1326,
  'W': 186.2132,
  'X': 110,
  'Y': 163.1760,
  'Z': 110
}

ion_type_props = {
  'b': { 
    'atoms_to_add': [], # Cterm O+ loses electron to form triple bond to C
    'atoms_to_sub': [],
    'is_nterm': True, 
  },
  'y': { 
    'atoms_to_add': ['H', 'H', 'O'], # H+ on Nterm NH, OH on Cterm CO
    'atoms_to_sub': [],
    'is_nterm': False,
  },
}


def parse_aa_masses(sequence, aa_mass, modified_masses=[]):
  masses = [aa_mass[aa] for aa in sequence]
  for modified in modified_masses:
    masses[modified['i']] = modified['mass']
  return masses


def ion_label(i, charge, ion_type):
  if charge == 1:
    return '%s%d' % (ion_type, i)
  return '%s%d(%d+)' % (ion_type, i, charge)


def calculate_neutral_mass(aa_masses, ion_type):
  props = ion_type_props[ion_type]
  mass = sum(aa_masses) 
  mass += sum([atom_mass[a] for a in props['atoms_to_add']])
  mass -= sum([atom_mass[a] for a in props['atoms_to_sub']])
  return mass


def calculate_mz(aa_masses, charge, ion_type):
  charged_mass = calculate_neutral_mass(aa_masses, ion_type)
  charged_mass += atom_mass['H']*charge 
  if charge == 1:
    return charged_mass
  else:
    return charged_mass/charge


def calculate_peaks(aa_masses, charge, ion_type):
  peaks = []
  for i in range(len(aa_masses)):
    if ion_type_props[ion_type]['is_nterm']:
      fragment = aa_masses[:i+1]
    else:
      fragment = aa_masses[i:]
    mz = calculate_mz(fragment, charge, ion_type)
    # round to 4 decimal places
    mz = int(mz*1E4)/1E4
    label = ion_label(len(fragment), charge, ion_type)
    peaks.append((mz, label))
  return peaks


def map_matched_ions(
    ion_type, sequence, peaks, mz_delta=0.8, modified_aa_masses=[],
    aa_mass=aa_monoisotopic_mass):
  aa_masses = parse_aa_masses(sequence, aa_mass, modified_aa_masses)
  ion = ion_type[0]
  pieces = ion_type.split('(')
  charge = pieces[1][0] if len(pieces) > 1 else 1
  theory_peaks = calculate_peaks(aa_masses, charge, ion)
  matched = []
  for theory_mz, label in theory_peaks:
    for mz, intensity in peaks:
      diff = theory_mz - mz
      error = int(diff/theory_mz*1E6)
      if abs(diff) <= mz_delta:
        matched.append([mz, intensity, label, diff, error])
        break
  return matched


if __name__ == "__main__":
  seq = 'MSAFLLTKR'
  spectrum = [[622.494, 2192.0],
   [615.666, 1811.0],
   [619.456, 1810.0],
   [628.245, 1772.0],
   [616.511, 1016.0],
   [614.583, 780.0],
   [629.003, 673.0],
   [490.215, 563.0],
   [606.378, 533.0],
   [563.797, 430.0],
   [610.871, 421.0],
   [472.861, 376.0],
   [607.103, 346.0],
   [584.56, 341.0],
   [605.67, 334.0],
   [620.333, 310.0],
   [375.222, 299.0],
   [625.311, 263.0],
   [488.298, 254.0],
   [786.349, 223.0],
   [899.377, 221.0],
   [506.51, 215.0],
   [881.571, 215.0],
   [601.8, 207.0],
   [514.999, 202.0],
   [598.419, 201.0],
   [608.138, 199.0],
   [593.698, 196.0],
   [759.359, 196.0],
   [530.555, 186.0],
   [623.222, 186.0],
   [602.825, 184.0],
   [542.091, 180.0],
   [261.063, 169.0],
   [627.002, 160.0],
   [913.689, 160.0],
   [592.707, 157.0],
   [825.581, 157.0],
   [471.826, 156.0],
   [505.835, 155.0],
   [609.168, 154.0],
   [768.284, 151.0],
   [563.027, 146.0],
   [570.712, 146.0],
   [580.803, 145.0],
   [979.473, 137.0],
   [596.779, 130.0],
   [541.113, 129.0],
   [777.839, 186.0],
   [550.083, 159.0]];
  delta_mass = 0.5
  modified_aa_masses = [{ 'i':3, 'mass':100.0 }]
  modified_aa_masses = []
  aa_masses = parse_aa_masses(seq, aa_monoisotopic_mass, modified_aa_masses)
  pprint(aa_masses)
  bion_peaks = calculate_peaks(aa_masses, 1, 'b')
  yion_peaks = calculate_peaks(aa_masses, 1, 'y')
  pprint(yion_peaks)
  peaks = []
  matched = map_matched_ions('y', seq, spectrum, delta_mass, modified_aa_masses, aa_monoisotopic_mass)
  matched = map_matched_ions('b', seq, spectrum, delta_mass, modified_aa_masses, aa_monoisotopic_mass)
  pprint(matched);




# -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint
import math
import json
import urllib
import logging


from parse import parse_string, save_data_dict


logger = logging.getLogger('mascot')


"""
Parser for Mascot dat files 

Some useful formulae:
 - identity_score = -10*Math.log10(1/qmatch)
 - qplughole = homology_score
 - significant if identity_score > homology_score

"""


def piece(s, splitter, i, j=None):
  if j is None:
    return str(s).split(splitter)[i]
  else:
    return str(s).split(splitter)[i:j]


def process_summary(l, scans):
  i = None
  if l.startswith('qmass'):
    lhs, rhs = piece(l, "=", 0, 2)
    i = int(lhs[5:])
    mass = rhs
    scans.setdefault(i, {})['theory_mass'] = float(mass)
  if l.startswith('qexp'):
    lhs, rhs = piece(l, "=", 0, 2)
    i = int(lhs[4:])
    r, c = rhs.split(",")
    scans.setdefault(i, {})['exp_mass_charge'] = float(r)
    scans.setdefault(i, {})['charge0'] = c
  if l.startswith('qplughole'):
    lhs, rhs = l.strip().split("=")
    i = int(lhs[9:])
    scans.setdefault(i, {})['homology'] = float(rhs)
  if l.startswith('qmatch'):
    lhs, rhs = l.strip().split("=")
    i = int(lhs[6:])
    val = float(rhs)
    if val == 0:
      identity = 0.0
    else:
      identity = 10.0*math.log10(float(val))
    scans.setdefault(i, {})['identity'] = identity
  if i is not None: 
    if 'matches' not in scans[i]:
      scans[i]['matches'] = []


def process_matches(l, scans, max_peptide_rank):
  query, data_str = piece(l, "=", 0, 2)
  if data_str.strip() == "-1":
    return
  if 'subs' in query:
    return

  entry, match = piece(query, "_", 0, 2)
  if max_peptide_rank is not None:
    i_match = int(match[1:])
    if i_match > max_peptide_rank:
      return

  query_pieces = query.split("_")
  scan_id = int(entry[1:])
  if 'terms' in query:
    if scan_id not in scans:
      return
    terms_list = data_str.split(":")
    sites = scans[scan_id]['matches'][-1]['proteins']
    for terms, site in zip(terms_list, sites):
      n_term, c_term = terms.split(",")
      site['n_term'] = n_term
      site['c_term'] = c_term
  elif len(query_pieces) == 2:
    d = data_str.split(",")
    try:
      data = {
        'skip': int(d[0]),
        'mass': float(d[1]),
        'mass_diff': float(d[2]),
        'sequence': d[4],
        'score': float(d[7]),
        'proteins': []
      }
    except:
      raise ValueError("Error reading: " + l)
    protein_sites_str = piece(l, ";", 1)
    protein_sites = protein_sites_str.split(",")

    i_match = len(scans[scan_id]['matches'])
    score = data['score']
    for protein_site in protein_sites:
      p = protein_site.split(":")
      seqid = p[0][1:-1]
      protein = {
        'seqid': seqid,
        'i': int(p[2]),
        'j': int(p[3]),
      }
      data['proteins'].append(protein)
    scans[scan_id]['matches'].append(data)


def process_proteins(l, proteins):
  i = l.index(",")
  lhs = l[:i]
  rhs = l[i:]
  words = lhs.split("=")
  seqid = words[0][1:-1]
  mass = float(words[1])
  name = rhs[1:-1]
  proteins[seqid] = {
    'mass': mass,
    'description': name
  }


def process_query(l, scan_id, scans):
  lhs, rhs = l.split("=")
  if lhs in scans[scan_id]:
    raise ValueError(
        "%s already in scans[%d]" % (lhs, scan_id))
  if lhs == 'title':
    scans[scan_id][lhs] = urllib.unquote(rhs)
  else:
    scans[scan_id][lhs] = parse_string(rhs)


def read_mascot_dat(fname, max_peptide_rank=1):
  scans = {}
  proteins = {}
  boundary_id = None
  section = None
  process_line = None
  for l in open(fname, 'rU'):
    if not l.strip():
      continue
    if not boundary_id:
      if l.startswith("Content-Type") and 'boundary' in l:
        boundary_id = "--" + piece(l, "=", -1)
        continue
    if not boundary_id:
      continue
    if boundary_id and l.startswith(boundary_id):
      section = None
      process_line = None
      continue
    if l.startswith("Content-Type"):
      word = piece(l, None, -1)
      name = piece(word, "=", -1)
      section = name[1:-1]
      if section == "summary":
        process_line = \
            lambda l: process_summary(l, scans)
      if section == "peptides":
        process_line = \
            lambda l: process_matches(l, scans, max_peptide_rank)
      if section == "proteins":
        process_line = \
            lambda l: process_proteins(l, proteins)
      if "query" in section:
        scan_id = int(section[5:])
        process_line = \
            lambda l: process_query(l, scan_id, scans)
      continue
    if process_line:
      process_line(l[:-1])
  return scans, proteins


def split_mascot_ion_str(s):
    "Parses an Ion entry string into a dictionary"
    pairs = [piece.split(':') for piece in s.split(',')]
    return [[float(x),float(y)] for x,y in pairs]


def load_mascot_dat_to_proteins(proteins, i_source, mascot_dat):
  peptide_by_match_id = {}
  for seqid in proteins:
    protein = proteins[seqid]
    for source in protein['sources']:
      for peptide in source['peptides']:
        if 'identity' in peptide['attr']:
          match_id = "%.2f%.2f%.2f%s" % \
              (peptide['attr']['score'],
               peptide['attr']['identity'],
               peptide['attr']['homology'],
               peptide['sequence'])
          peptide_by_match_id[match_id] = peptide
  scans, mascot_proteins = read_mascot_dat(mascot_dat)
  n_match = 0
  for scan in scans.values():
    for match in scan['matches']:
      match_id = "%.2f%.2f%.2f%s" % \
          (match['score'],
           scan['identity'],
           scan['homology'],
           match['sequence'])
      if match_id in peptide_by_match_id:
        n_match += 1
        peptide = peptide_by_match_id[match_id]
        peptide['spectrum'] = split_mascot_ion_str(scan['Ions1'])
  logger.info('%s: matched %d pepXML to %d mascot PSM' % \
      (mascot_dat, len(peptide_by_match_id), n_match))


if __name__ == '__main__':
  scans, proteins = read_mascot_dat('../example/mascot/F022045.dat')
  save_data_dict(scans, '../example/mascot/scans.dump')
  save_data_dict(proteins,'../example/mascot/proteins.dump')



 # -*- coding: utf-8 -*-
from __future__ import print_function
from pprint import pprint
import os
import json
import math
import logging

logger = logging.getLogger('maxquant')

import numpy 

import parse
import proteins as parse_proteins

"""
Parser for Maxquant summary text files
"""


def get_labeled_spectrum(scan):
  masses = parse.splitter(scan['masses'], float)
  intensities = parse.splitter(scan['intensities'], float)
  labels = parse.splitter(scan['matches'])
  return [x for x in zip(masses, intensities, labels)]


def transfer_attrs(source_dict, target_dict, parse_list):
  for key, convert_fn in parse_list:
    if key in source_dict:
      target_dict[key] = convert_fn(source_dict[key])


def float_or_none(float_str):
  if float_str == "":
    return None
  return float(float_str)


evidence_parse_list = [
  ('intensity', float_or_none),
  ('experiment', lambda s: s.lower()),
  ('ratio h/l', float_or_none),
  ('ratio h/l normalized', float_or_none),
]

peptide_parse_list = [
  ('ratio h/l variability [%]', float_or_none),
]

protein_parse_list = [
  ('ratio h/l', float),
  ('ratio h/l normalized', float),
  ('ratio h/l variability [%]', float),
]

scan_parse_list = [
  ('scan number', int),
  ('m/z', float),
  ('labeling state', int),
  ('retention time', float),
  ('pep', float),
]


def change_key(data, old_key, new_key):
  if not old_key in data or old_key == new_key:
    return
  data[new_key] = data[old_key]
  del data[old_key]


def read_tsv(tsv_txt):
  """
  Reads a title top TSV file, converts all keys to lower case
  and returns a list of dictionaries.
  """
  f = open(tsv_txt, "UR")
  titles = [w.lower() for w in split_tab(f.readline())]
  results = []
  for line in f.readlines():
    group = {}
    for key, val in zip(titles, split_tab(line)):
      group[key] = parse_string(val)
    results.append(group)
  return results


def get_proteins_and_sources(in_dir, is_leu_ile_isomeric=False,):

  evidence_fname = os.path.join(in_dir, 'evidence.txt')
  logger.info('Loading evidence file: ' + evidence_fname)
  evidence_iter = parse.read_tsv(evidence_fname)
  evidence_dict = { int(e['id']):e for e in evidence_iter }

  sources_set = set(e['raw file'] for e in evidence_dict.values())
  sources = [str(s) for s in sorted(sources_set)]
  i_sources = {source:k for k, source in enumerate(sources)}

  protein_group_fname = os.path.join(in_dir, 'proteinGroups.txt')
  logger.info('Loading protein groups: ' + protein_group_fname)
  proteins = {}
  protein_by_group_id = {}
  for protein_group in parse.read_tsv(protein_group_fname):
    group_id = protein_group['id']
    protein = {
      'description': '',
      'attr': { 
        'group_id': group_id,
        'other_seqids': [],
      },
      'sources': [{ 'peptides': [] } for k in range(len(i_sources))],
    }
    transfer_attrs(protein_group, protein['attr'], protein_parse_list)

    seqids = parse.splitter(protein_group['protein ids'])
    proteins[seqids[0]] = protein
    protein['attr']['seqid'] = seqids[0]
    protein['attr']['other_seqids'] = seqids[1:]
    protein_by_group_id[group_id] = protein

  peptides_fname = os.path.join(in_dir, 'peptides.txt')
  logger.info('Loading peptides file: ' + peptides_fname)
  peptides_iter = parse.read_tsv(peptides_fname)
  peptides = { int(p['id']):p for p in peptides_iter }

  scans_fname = os.path.join(in_dir, 'msms.txt')
  logger.info('Loading scans and matching: ' + scans_fname)
  i_scan = 0
  for scan in parse.read_tsv(scans_fname):
    scan_id = int(scan['id'])
    i_scan += 1
    if i_scan % 5000 == 0:
      logger.info("{} scans processed".format(i_scan))
    evidence_id = int(scan['evidence id'])
    evidence = evidence_dict[evidence_id]

    peptide_id = int(scan['peptide id'])
    peptide = peptides[peptide_id]
    for group_id in parse.splitter(str(scan['protein group ids'])):
      new_peptide = {
        'sequence': scan['sequence'],
        'spectrum': get_labeled_spectrum(scan),
        'attr' : {
          'modifications': [],
          'mq_scan_id': scan_id,
          'is_unique': peptide['unique (groups)'] == 'yes',
        }
      }
      transfer_attrs(scan, new_peptide['attr'], scan_parse_list)
      transfer_attrs(evidence, new_peptide['attr'], evidence_parse_list)
      transfer_attrs(peptide, new_peptide['attr'], peptide_parse_list)
      change_key(new_peptide['attr'], 'scan number', 'scan_id')
      change_key(new_peptide['attr'], 'retention time', 'retention_time')
      
      protein = protein_by_group_id[int(group_id)]
      i_source = i_sources[evidence['raw file']]
      protein['sources'][i_source]['peptides'].append(new_peptide)

  parse_proteins.count_peptides(proteins)
  
  return proteins, sources


def scale_log(r, max_v):
  v = math.log(r)
  scale_v = v/max_v
  if scale_v > 1.0:
    scale_v = 1.0
  if scale_v < -1.0:
    scale_v = -1.0
  return scale_v


def calculate_ratio_intensities(
    proteins, max_ratio=2.0, ratio_key='ratio h/l normalized haha'):
  for seqid in proteins.keys():
    for source in proteins[seqid]['sources']:
       for peptide in source['peptides']:
        peptide['intensity'] = ""
        if ratio_key in peptide['attr']:
          ratio = peptide['attr'][ratio_key]
          if ratio is None or math.isnan(ratio) or ratio < 0:
            peptide['intensity'] = scale_log(ratio, max_ratio)


def calculate_lfq_ratio_intensities(
    proteins, experiment1, experiment2, max_ratio=2.0):
  for seqid in proteins:
    protein = proteins[seqid]
    peptide_intensities1 = []
    peptide_intensities2 = []
    for source in protein['sources']:
      peptides_by_seq = {}
      for peptide in source['peptides']:
        seq = peptide['sequence']
        peptides_by_seq.setdefault(seq, []).append(peptide)
      for seq, peptides in peptides_by_seq.items():
        intensities = {}
        for peptide in peptides:
          experiment = peptide['attr']['experiment']
          if not peptide['attr']['intensity']: 
            continue
          if experiment not in intensities:
            intensities[experiment] = []
          intensities[experiment].append(float(peptide['attr']['intensity']))
        if experiment1 in intensities and experiment2 in intensities:
          intensity1 = numpy.mean(intensities[experiment1])
          intensity2 = numpy.mean(intensities[experiment2])
          std1 = numpy.std(intensities[experiment1])
          std2 = numpy.std(intensities[experiment2])
          ratio = intensity1/intensity2
          std = ratio*numpy.sqrt((std1/intensity1)**2 + (std2/intensity2)**2)
          intensity = scale_log(ratio, max_ratio)
          std = numpy.round(std, 4)
          ratio = numpy.round(ratio, 4)
          peptide_intensities1.extend(intensities[experiment1])
          peptide_intensities2.extend(intensities[experiment2])
        elif experiment1 in intensities:
          ratio = float("inf")
          intensity = 2*max_ratio
          std = 0
          peptide_intensities1.extend(intensities[experiment1])
        elif experiment2 in intensities:
          ratio = 0.0
          intensity = -2*max_ratio
          std = 0
          peptide_intensities2.extend(intensities[experiment2])
        else:
          # neither experiment1 or experiment2 found in experiment column
          ratio = None
        for peptide in peptides:
          peptide['attr']['ratio'] = ratio
          peptide['intensity'] = intensity
          peptide['attr']['ratio_var'] = std
    sum2 = numpy.sum(peptide_intensities2)
    sum1 = numpy.sum(peptide_intensities1)
    if sum2 > 0.0:
      group_ratio = sum1/sum2
    else:
      group_ratio = float('inf')
    protein['attr']['ratio'] = group_ratio







#!/usr/bin/env python
from setuptools import setup

setup(
    name='peptagram',
    version='0.2',
    author='Bosco Ho',
    author_email='boscoh@gmail.com',
    url='http://boscoh.github.io/peptagram',
    description='python scripts to generate HTML5/javascript visualization for proteomics data',
    long_description="Docs at http://boscoh.github.io/peptagram",
    license='BSD',
    include_package_data = True,
    install_requires=[
        'uniprot',
        'pymzml',
    ],
    packages=['peptagram',],
    scripts=[],
)